package uk.ac.ed.pguaglia.real.db;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonMappingException;

import uk.ac.ed.pguaglia.real.db.engine.BaseTable;
import uk.ac.ed.pguaglia.real.db.engine.DataValue.DataType;
import uk.ac.ed.pguaglia.real.lang.Expression;

public class Schema {

	private Map<String,BaseTable> tables;
	private Map<String,Expression> views;
	private Map<String,Set<String>> isUsedBy;

	public Schema() {
		tables = new HashMap<>();
		views = new HashMap<>();
		isUsedBy = new HashMap<>();
	}

	public static Schema fromJSON(File src) throws JsonParseException, JsonMappingException, IOException {
		return SchemaMapper.getInstance().readValue(src, Schema.class);
	}

	public File getTableDatafile(String name) throws IOException {
		return tables.get(name).datafile().toFile();
	}

	public boolean hasTable(String name) {
		return tables.containsKey(name);
	}

	public boolean hasView(String name) {
		return views.containsKey(name);
	}

	public List<String> getTableAttributes(String name) {
		return tables.get(name).attributes();
	}

	public Expression getViewDefinition(String name) {
		return views.get(name);
	}

	public void addTable(String name, BaseTable table) throws Exception {
		if (tables.containsKey(name)) {
			throw new Exception(String.format("ERROR: Table %s already exists", name));
		}
		tables.put(name, table);
	}

	public void dropTable(String name) throws Exception {
		if (isUsedBy.containsKey(name)) {
			String msg = String.format("ERROR: Table %s is used by views %s", name,
					String.join(", ", isUsedBy.get(name)));
			throw new Exception(msg);
		}
		tables.remove(name);
	}

	public Expression addView(String name, Expression def) throws SchemaException {
		if (views.containsKey(name)) {
			String msg = String.format("View \"%s\" already defined", name);
			throw new SchemaException(msg, def, this);
		}
		if (tables.containsKey(name)) {
			String msg = String.format("A table with name \"%s\" already exists", name);
			throw new SchemaException(msg, def, this);
		}
		views.put(name, def);
		updateDeps(name, def);
		return def;
	}

	public Expression dropView(String name) throws SchemaException {
		if (isUsedBy.containsKey(name)) {
			String msg = String.format("View \"%s\" is used by \"%s\"", name, String.join("\", \"", isUsedBy.get(name)));
			throw new SchemaException(msg, null, this);
		}
		Iterator<String> it = isUsedBy.keySet().iterator();
		while (it.hasNext()) {
			String base = it.next();
			Set<String> deps = isUsedBy.get(base);
			deps.remove(name);
			if (deps.isEmpty()) {
				it.remove();
			}
		}
		Expression def = views.remove(name);
		return def;
	}

	private void updateDeps(String name, Expression expr) {
		for (String base : expr.getBaseNames()) {
			if (isUsedBy.containsKey(base)) {
				isUsedBy.get(base).add(name);
			} else {
				Set<String> deps = new HashSet<>();
				deps.add(name);
				isUsedBy.put(base, deps);
			}
		}
	}

	public Set<String> getViewNames() {
		return views.keySet();
	}

	public Set<String> getTableNames() {
		return tables.keySet();
	}

	public DataType getDataType(String tbl, String attr) {
		return tables.get(tbl).getDataType(attr);
	}

	public Set<String> getTableNames(Expression expr) throws SchemaException {
		Set<String> base = new HashSet<>();
		for (String name : expr.getBaseNames()) {
			if (this.hasView(name)) {
				base.addAll(this.getTableNames(this.getViewDefinition(name)));
			} else if (this.hasTable(name)) {
				base.add(name);
			} else {
				String msg = String.format("No such table \"%s\"", name);
				throw new SchemaException(msg, expr, this);
			}
		}
		return base;
	}
}
